@@ -45,6 +45,7 @@
 #include "gdm-manager-glue.h"
 #include "gdm-display-store.h"
 #include "gdm-display-factory.h"
+#include "gdm-launch-environment.h"
 #include "gdm-local-display.h"
 #include "gdm-local-display-factory.h"
 #include "gdm-session.h"
@@ -52,6 +53,7 @@
 #include "gdm-settings-direct.h"
 #include "gdm-settings-keys.h"
 #include "gdm-xdmcp-display-factory.h"
+#include "gdm-xdmcp-chooser-display.h"
 
 #define GDM_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GDM_TYPE_MANAGER, GdmManagerPrivate))
 
@@ -80,12 +82,10 @@ struct GdmManagerPrivate
         GHashTable             *transient_sessions;
         GHashTable             *open_reauthentication_requests;
         gboolean                xdmcp_enabled;
-        GCancellable           *cancellable;
 
         gboolean                started;
         gboolean                show_local_greeter;
 
-        GDBusProxy               *bus_proxy;
         GDBusConnection          *connection;
         GDBusObjectManagerServer *object_manager;
 
@@ -121,12 +121,12 @@ static void     gdm_manager_class_init  (GdmManagerClass *klass);
 static void     gdm_manager_init        (GdmManager      *manager);
 static void     gdm_manager_dispose     (GObject         *object);
 
-static GdmSession *create_embryonic_user_session_for_display (GdmManager *manager,
-                                                              GdmDisplay *display,
-                                                              uid_t       allowed_user);
-
+static GdmSession *create_user_session_for_display (GdmManager *manager,
+                                                    GdmDisplay *display,
+                                                    uid_t       allowed_user);
 static void     start_user_session (GdmManager                *manager,
                                     StartUserSessionOperation *operation);
+static void     clean_user_session (GdmSession *session);
 
 static gpointer manager_object = NULL;
 
@@ -294,37 +294,6 @@ is_login_session (GdmManager  *self,
         return TRUE;
 }
 
-static gboolean
-activate_session_id (GdmManager *manager,
-                     const char *seat_id,
-                     const char *session_id)
-{
-        GError *error = NULL;
-        GVariant *reply;
-
-        reply = g_dbus_connection_call_sync (manager->priv->connection,
-                                             "org.freedesktop.login1",
-                                             "/org/freedesktop/login1",
-                                             "org.freedesktop.login1.Manager",
-                                             "ActivateSessionOnSeat",
-                                             g_variant_new ("(ss)", session_id, seat_id),
-                                             NULL, /* expected reply */
-                                             G_DBUS_CALL_FLAGS_NONE,
-                                             -1,
-                                             NULL,
-                                             &error);
-        if (reply == NULL) {
-                g_debug ("GdmManager: logind 'ActivateSessionOnSeat' %s raised:\n %s\n\n",
-                         g_dbus_error_get_remote_error (error), error->message);
-                g_error_free (error);
-                return FALSE;
-        }
-
-        g_variant_unref (reply);
-
-        return TRUE;
-}
-
 static gboolean
 session_unlock (GdmManager *manager,
                 const char *ssid)
@@ -621,7 +590,7 @@ switch_to_compatible_user_session (GdmManager *manager,
         if (existing_session != NULL) {
                 ssid_to_activate = gdm_session_get_session_id (existing_session);
                 if (seat_id != NULL) {
-                        res = activate_session_id (manager, seat_id, ssid_to_activate);
+                        res = gdm_activate_session_by_id (manager->priv->connection, seat_id, ssid_to_activate);
                         if (! res) {
                                 g_debug ("GdmManager: unable to activate session: %s", ssid_to_activate);
                                 goto out;
@@ -650,13 +619,13 @@ get_display_for_user_session (GdmSession *session)
 }
 
 static GdmSession *
-get_embryonic_user_session_for_display (GdmDisplay *display)
+get_user_session_for_display (GdmDisplay *display)
 {
         if (display == NULL) {
                 return NULL;
         }
 
-        return g_object_get_data (G_OBJECT (display), "gdm-embryonic-user-session");
+        return g_object_get_data (G_OBJECT (display), "gdm-user-session");
 }
 
 static gboolean
@@ -675,6 +644,11 @@ add_session_record (GdmManager    *manager,
         display_device = NULL;
 
         username = gdm_session_get_username (session);
+
+        if (username == NULL) {
+                goto out;
+        }
+
         g_object_get (G_OBJECT (session),
                       "display-name", &display_name,
                       "display-hostname", &hostname,
@@ -719,8 +693,8 @@ out:
 }
 
 static GdmSession *
-get_user_session_for_display (GdmManager *self,
-                              GdmDisplay *display)
+find_user_session_for_display (GdmManager *self,
+                               GdmDisplay *display)
 {
 
         GList *node = self->priv->user_sessions;
@@ -780,7 +754,7 @@ gdm_manager_handle_register_display (GdmDBusManager        *manager,
                 }
         }
 
-        session = get_user_session_for_display (self, display);
+        session = find_user_session_for_display (self, display);
 
         if (session != NULL) {
                 GPid pid;
@@ -820,7 +794,7 @@ gdm_manager_handle_open_session (GdmDBusManager        *manager,
         const char       *sender;
         GDBusConnection  *connection;
         GdmDisplay       *display = NULL;
-        GdmSession       *session;
+        GdmSession       *session = NULL;
         const char       *address;
         GPid              pid = 0;
         uid_t             uid = (uid_t) -1;
@@ -841,14 +815,35 @@ gdm_manager_handle_open_session (GdmDBusManager        *manager,
                 return TRUE;
         }
 
-        session = get_embryonic_user_session_for_display (display);
+#ifdef HAVE_LIBXDMCP
+        if (GDM_IS_XDMCP_CHOOSER_DISPLAY (display)) {
+                GdmLaunchEnvironment *launch_environment;
 
-        if (gdm_session_is_running (session)) {
-                g_dbus_method_invocation_return_error_literal (invocation,
-                                                               G_DBUS_ERROR,
-                                                               G_DBUS_ERROR_ACCESS_DENIED,
-                                                               _("Can only be called before user is logged in"));
-                return TRUE;
+                g_object_get (display, "launch-environment", &launch_environment, NULL);
+
+                if (launch_environment != NULL) {
+                        session = gdm_launch_environment_get_session (launch_environment);
+                }
+
+                if (session == NULL) {
+                        g_dbus_method_invocation_return_error_literal (invocation,
+                                                                       G_DBUS_ERROR,
+                                                                       G_DBUS_ERROR_ACCESS_DENIED,
+                                                                       _("Chooser session unavailable"));
+                        return TRUE;
+                }
+        }
+#endif
+        if (session == NULL) {
+                session = get_user_session_for_display (display);
+
+                if (gdm_session_is_running (session)) {
+                        g_dbus_method_invocation_return_error_literal (invocation,
+                                                                       G_DBUS_ERROR,
+                                                                       G_DBUS_ERROR_ACCESS_DENIED,
+                                                                       _("Can only be called before user is logged in"));
+                        return TRUE;
+                }
         }
 
         allowed_user = gdm_session_get_allowed_user (session);
@@ -1102,7 +1097,7 @@ gdm_manager_handle_open_reauthentication_channel (GdmDBusManager        *manager
                                                          seat_id,
                                                          NULL);
         } else {
-                session = get_embryonic_user_session_for_display (display);
+                session = get_user_session_for_display (display);
         }
 
         if (session != NULL && gdm_session_is_running (session)) {
@@ -1320,7 +1315,7 @@ set_up_automatic_login_session (GdmManager *manager,
         /* 0 is root user; since the daemon talks to the session object
          * directly, itself, for automatic login
          */
-        session = create_embryonic_user_session_for_display (manager, display, 0);
+        session = create_user_session_for_display (manager, display, 0);
 
         g_object_get (G_OBJECT (display),
                       "is-initial", &is_initial,
@@ -1335,6 +1330,26 @@ set_up_automatic_login_session (GdmManager *manager,
         gdm_session_start_conversation (session, "gdm-autologin");
 }
 
+static void
+set_up_chooser_session (GdmManager *manager,
+                        GdmDisplay *display)
+{
+        const char *allowed_user;
+        struct passwd *passwd_entry;
+
+        allowed_user = get_username_for_greeter_display (manager, display);
+
+        if (!gdm_get_pwent_for_name (allowed_user, &passwd_entry)) {
+                g_warning ("GdmManager: couldn't look up username %s",
+                           allowed_user);
+                gdm_display_unmanage (display);
+                gdm_display_finish (display);
+                return;
+        }
+
+        gdm_display_start_greeter_session (display);
+}
+
 static void
 set_up_greeter_session (GdmManager *manager,
                         GdmDisplay *display)
@@ -1352,7 +1367,7 @@ set_up_greeter_session (GdmManager *manager,
                 return;
         }
 
-        create_embryonic_user_session_for_display (manager, display, passwd_entry->pw_uid);
+        create_user_session_for_display (manager, display, passwd_entry->pw_uid);
         gdm_display_start_greeter_session (display);
 }
 
@@ -1403,15 +1418,26 @@ set_up_session (GdmManager *manager,
         ActUserManager *user_manager;
         ActUser *user;
         gboolean loaded;
+        gboolean is_initial_display = FALSE;
         gboolean autologin_enabled = FALSE;
         char *username = NULL;
 
-        if (!manager->priv->ran_once && display_is_on_seat0 (display))
+        g_object_get (G_OBJECT (display), "is-initial", &is_initial_display, NULL);
+
+        if (!manager->priv->ran_once && is_initial_display)
                 autologin_enabled = get_automatic_login_details (manager, &username);
 
         if (!autologin_enabled) {
-                set_up_greeter_session (manager, display);
                 g_free (username);
+
+#ifdef HAVE_LIBXDMCP
+                if (GDM_IS_XDMCP_CHOOSER_DISPLAY (display)) {
+                        set_up_chooser_session (manager, display);
+                        return;
+                }
+#endif
+
+                set_up_greeter_session (manager, display);
                 return;
         }
 
@@ -1446,8 +1472,6 @@ greeter_display_started (GdmManager *manager,
         }
 
         maybe_start_pending_initial_login (manager, display);
-
-        manager->priv->ran_once = TRUE;
 }
 
 static void
@@ -1457,6 +1481,7 @@ on_display_status_changed (GdmDisplay *display,
 {
         int         status;
         int         display_number = -1;
+        char       *session_type = NULL;
 #ifdef WITH_PLYMOUTH
         gboolean    display_is_local = FALSE;
         gboolean    quit_plymouth = FALSE;
@@ -1467,7 +1492,10 @@ on_display_status_changed (GdmDisplay *display,
         quit_plymouth = display_is_local && manager->priv->plymouth_is_running;
 #endif
 
-        g_object_get (display, "x11-display-number", &display_number, NULL);
+        g_object_get (display,
+                      "x11-display-number", &display_number,
+                      "session-type", &session_type,
+                      NULL);
 
         status = gdm_display_get_status (display);
 
@@ -1500,6 +1528,9 @@ on_display_status_changed (GdmDisplay *display,
                         }
 #endif
 
+                        if (status == GDM_DISPLAY_FINISHED || g_strcmp0 (session_type, "x11") == 0) {
+                                manager->priv->ran_once = TRUE;
+                        }
                         maybe_start_pending_initial_login (manager, display);
                         break;
                 default:
@@ -1594,6 +1625,11 @@ create_display_for_user_session (GdmManager *self,
         gdm_display_store_add (self->priv->display_store,
                                display);
         g_object_set_data (G_OBJECT (session), "gdm-display", display);
+        g_object_set_data_full (G_OBJECT (display),
+                                "gdm-user-session",
+                                g_object_ref (session),
+                                (GDestroyNotify)
+                                clean_user_session);
 }
 
 static gboolean
@@ -1605,6 +1641,7 @@ on_start_user_session (StartUserSessionOperation *operation)
         gboolean doing_initial_setup = FALSE;
         gboolean starting_user_session_right_away = TRUE;
         GdmDisplay *display;
+        const char *session_id;
 
         g_debug ("GdmManager: start or jump to session");
 
@@ -1629,14 +1666,19 @@ on_start_user_session (StartUserSessionOperation *operation)
 
         g_object_get (G_OBJECT (display), "doing-initial-setup", &doing_initial_setup, NULL);
 
+        session_id = gdm_session_get_conversation_session_id (operation->session,
+                                                              operation->service_name);
+
         if (gdm_session_get_display_mode (operation->session) == GDM_SESSION_DISPLAY_MODE_REUSE_VT) {
                 /* In this case, the greeter's display is morphing into
                  * the user session display. Kill the greeter on this session
                  * and let the user session follow the same display. */
                 gdm_display_stop_greeter_session (display);
-                g_object_set (G_OBJECT (display), "session-class", "user", NULL);
+                g_object_set (G_OBJECT (display),
+                                "session-class", "user",
+                                "session-id", session_id,
+                                NULL);
         } else {
-                const char *session_id;
                 uid_t allowed_uid;
 
                 g_object_ref (display);
@@ -1659,11 +1701,11 @@ on_start_user_session (StartUserSessionOperation *operation)
 
                 /* The user session is going to follow the session worker
                  * into the new display. Untie it from this display and
-                 * create a new embryonic session for a future user login. */
+                 * create a new session for a future user login. */
                 allowed_uid = gdm_session_get_allowed_user (operation->session);
-                g_object_set_data (G_OBJECT (display), "gdm-embryonic-user-session", NULL);
+                g_object_set_data (G_OBJECT (display), "gdm-user-session", NULL);
                 g_object_set_data (G_OBJECT (operation->session), "gdm-display", NULL);
-                create_embryonic_user_session_for_display (operation->manager, display, allowed_uid);
+                create_user_session_for_display (operation->manager, display, allowed_uid);
 
                 if (g_strcmp0 (operation->service_name, "gdm-autologin") == 0) {
                         /* remove the unused prepared greeter display since we're not going
@@ -1673,8 +1715,6 @@ on_start_user_session (StartUserSessionOperation *operation)
                 }
 
                 /* Give the user session a new display object for bookkeeping purposes */
-                session_id = gdm_session_get_conversation_session_id (operation->session,
-                                                                      operation->service_name);
                 create_display_for_user_session (operation->manager,
                                                  operation->session,
                                                  session_id);
@@ -1792,9 +1832,10 @@ remove_user_session (GdmManager *manager,
 static void
 on_session_start_failed (GdmSession *session,
                          const char *service_name,
+                         const char *message,
                          GdmManager *manager)
 {
-        g_debug ("GdmManager: session failed to start");
+        g_debug ("GdmManager: session failed to start: %s", message);
         remove_user_session (manager, session);
 }
 
@@ -1838,6 +1879,26 @@ on_session_reauthenticated (GdmSession *session,
                             GdmManager *manager)
 {
         gboolean fail_if_already_switched = FALSE;
+
+        if (gdm_session_get_display_mode (session) == GDM_SESSION_DISPLAY_MODE_REUSE_VT) {
+                const char *seat_id;
+                char *session_id;
+
+                seat_id = gdm_session_get_display_seat_id (session);
+                if (gdm_get_login_window_session_id (seat_id, &session_id)) {
+                        GdmDisplay *display = gdm_display_store_find (manager->priv->display_store,
+                                                                      lookup_by_session_id,
+                                                                      (gpointer) session_id);
+
+                        if (display != NULL) {
+                                gdm_display_stop_greeter_session (display);
+                                gdm_display_unmanage (display);
+                                gdm_display_finish (display);
+                        }
+                        g_free (session_id);
+                }
+        }
+
         /* There should already be a session running, so jump to its
          * VT. In the event we're already on the right VT, (i.e. user
          * used an unlock screen instead of a user switched login screen),
@@ -1930,25 +1991,7 @@ on_session_client_disconnected (GdmSession   *session,
                                 GPid          pid_of_client,
                                 GdmManager   *manager)
 {
-        GdmDisplay *display;
-        gboolean display_is_local;
-
         g_debug ("GdmManager: client disconnected");
-
-        display = get_display_for_user_session (session);
-
-        if (display == NULL) {
-                return;
-        }
-
-        g_object_get (G_OBJECT (display),
-                      "is-local", &display_is_local,
-                      NULL);
-
-        if ( ! display_is_local && gdm_session_is_running (session)) {
-                gdm_display_unmanage (display);
-                gdm_display_finish (display);
-        }
 }
 
 typedef struct
@@ -2081,16 +2124,16 @@ on_session_reauthentication_started (GdmSession *session,
 }
 
 static void
-clean_embryonic_user_session (GdmSession *session)
+clean_user_session (GdmSession *session)
 {
         g_object_set_data (G_OBJECT (session), "gdm-display", NULL);
         g_object_unref (session);
 }
 
 static GdmSession *
-create_embryonic_user_session_for_display (GdmManager *manager,
-                                           GdmDisplay *display,
-                                           uid_t       allowed_user)
+create_user_session_for_display (GdmManager *manager,
+                                 GdmDisplay *display,
+                                 uid_t       allowed_user)
 {
         GdmSession *session;
         gboolean    display_is_local = FALSE;
@@ -2100,7 +2143,7 @@ create_embryonic_user_session_for_display (GdmManager *manager,
         char       *display_auth_file = NULL;
         char       *display_seat_id = NULL;
         char       *display_id = NULL;
-#ifdef ENABLE_WAYLAND_SUPPORT
+#if defined(ENABLE_WAYLAND_SUPPORT) && defined(ENABLE_USER_DISPLAY_SERVER)
         char       *display_session_type = NULL;
         gboolean    greeter_is_wayland;
 #endif
@@ -2112,7 +2155,7 @@ create_embryonic_user_session_for_display (GdmManager *manager,
                       "remote-hostname", &remote_hostname,
                       "x11-authority-file", &display_auth_file,
                       "seat-id", &display_seat_id,
-#ifdef ENABLE_WAYLAND_SUPPORT
+#if defined(ENABLE_WAYLAND_SUPPORT) && defined(ENABLE_USER_DISPLAY_SERVER)
                       "session-type", &display_session_type,
 #endif
                       NULL);
@@ -2190,12 +2233,12 @@ create_embryonic_user_session_for_display (GdmManager *manager,
                           manager);
         g_object_set_data (G_OBJECT (session), "gdm-display", display);
         g_object_set_data_full (G_OBJECT (display),
-                                "gdm-embryonic-user-session",
+                                "gdm-user-session",
                                 g_object_ref (session),
                                 (GDestroyNotify)
-                                clean_embryonic_user_session);
+                                clean_user_session);
 
-#ifdef ENABLE_WAYLAND_SUPPORT
+#if defined(ENABLE_WAYLAND_SUPPORT) && defined(ENABLE_USER_DISPLAY_SERVER)
         greeter_is_wayland = g_strcmp0 (display_session_type, "wayland") == 0;
         g_object_set (G_OBJECT (session), "ignore-wayland", !greeter_is_wayland, NULL);
 #endif
@@ -2308,6 +2351,13 @@ gdm_manager_start (GdmManager *manager)
 #ifdef HAVE_LIBXDMCP
         /* Accept remote connections */
         if (manager->priv->xdmcp_enabled) {
+#ifdef WITH_PLYMOUTH
+                /* Quit plymouth if xdmcp is the only display */
+                if (!manager->priv->show_local_greeter && manager->priv->plymouth_is_running) {
+                        plymouth_quit_with_transition ();
+                        manager->priv->plymouth_is_running = FALSE;
+                }
+#endif
                 if (manager->priv->xdmcp_factory != NULL) {
                         g_debug ("GdmManager: Accepting XDMCP connections...");
                         gdm_display_factory_start (GDM_DISPLAY_FACTORY (manager->priv->xdmcp_factory));
@@ -2331,7 +2381,7 @@ register_manager (GdmManager *manager)
         if (manager->priv->connection == NULL) {
                 g_critical ("error getting system bus: %s", error->message);
                 g_error_free (error);
-                exit (1);
+                exit (EXIT_FAILURE);
         }
 
         object_server = g_dbus_object_manager_server_new (GDM_MANAGER_DISPLAYS_PATH);
@@ -2346,7 +2396,7 @@ register_manager (GdmManager *manager)
                             GDM_MANAGER_PATH,
                             error->message);
                 g_error_free (error);
-                exit (1);
+                exit (EXIT_FAILURE);
         }
 
         return TRUE;
@@ -2545,6 +2595,7 @@ finish_display (const char *id,
                 GdmDisplay *display,
                 GdmManager *manager)
 {
+        gdm_display_stop_greeter_session (display);
         if (gdm_display_get_status (display) == GDM_DISPLAY_MANAGED)
                 gdm_display_unmanage (display);
         gdm_display_finish (display);
